MAPFILE = "mapfile"
INTERFACE = "insterface"
IDENTIFIER = "identifier"

PREFIS = "prefix"
SIMULATED =  "simulated"
IGNORED = "ignored"
ADDRESS = "address"
SIMADDR = "simaddr"


PREFIX = "prefix"
FITS_PREFIX = "fits_prefix"

CTRL_CONFIG = "ctrl_config"
INITIALISATION = "initialisation"

SERVER_ID = 'server_id'
DEVICES = 'devices'
CFGFILE = 'cfgfile'
NAMESPACE = "namespace"

STATE = "state"
SUBSTATE = "substate"
STATUS = "status"

CFG = "cfg"
STAT = "stat"
RPC = "rpc"
TYPE = "type"
